@@ -103,12 +103,22 @@ class TestPandasConversions(object):
         with localconverter(default_converter + rpyp.converter) as cv:
             rp_df = robjects.conversion.converter_ctx.get().py2rpy(pd_df)
         assert tuple(rp_df.names) == ('the one', 'the other')
-        
+
+    def test_dataframe_warns_duplicated_index(self):
+        pd_df = pandas.DataFrame({'x': [1, 2]}, index=['a', 'a'])
+        with pytest.warns(
+            UserWarning,
+            match='DataFrame contains duplicated elements in the index'
+        ) as record:
+            with localconverter(default_converter + rpyp.converter) as cv:
+                robjects.conversion.converter_ctx.get().py2rpy(pd_df)
+        assert len(record) == 1
+
     def test_series(self):
         Series = pandas.core.series.Series
         s = Series(numpy.random.randn(5), index=['a', 'b', 'c', 'd', 'e'])
         with localconverter(default_converter + rpyp.converter) as cv:
-            rp_s = robjects.conversion.get_conversion().py2rpy(s)
+            rp_s = robjects.conversion.converter_ctx.get().py2rpy(s)
         assert isinstance(rp_s, rinterface.FloatSexpVector)
 
     @pytest.mark.parametrize('dtype',
@@ -183,6 +193,8 @@ class TestPandasConversions(object):
             with pytest.raises(ValueError):
                 rp_s = robjects.conversion.converter_ctx.get().py2rpy(s)
 
+    @pytest.mark.skipif(not has_pandas,
+                        reason='pandas must be installed.')
     @pytest.mark.parametrize(
         'data',
         ([True, False, True],
@@ -191,12 +203,16 @@ class TestPandasConversions(object):
     @pytest.mark.parametrize(
         'dtype', [bool, pandas.BooleanDtype() if has_pandas else None]
     )
-    def test_series_obj_bool(self, data, dtype):
-        Series = pandas.core.series.Series
-        s = Series(data, index=['a', 'b', 'c'], dtype=dtype)
+    @pytest.mark.parametrize(
+        'constructor,wrapcheck',
+        [(pandas.Series, lambda x: x),
+         (pandas.DataFrame, lambda x: x[0])]
+    )
+    def test_series_obj_bool(self, data, dtype, constructor, wrapcheck):
+        s = constructor(data, index=['a', 'b', 'c'], dtype=dtype)
         with localconverter(default_converter + rpyp.converter) as cv:
             rp_s = robjects.conversion.converter_ctx.get().py2rpy(s)
-        assert isinstance(rp_s, rinterface.BoolSexpVector)
+        assert isinstance(wrapcheck(rp_s), rinterface.BoolSexpVector)
 
 
     @pytest.mark.parametrize(
@@ -226,7 +242,7 @@ class TestPandasConversions(object):
     def test_series_issue264(self):
         Series = pandas.core.series.Series
         s = Series(('a', 'b', 'c', 'd', 'e'),
-                   index=pandas.Int64Index([0,1,2,3,4]))
+                   index=pandas.Index([0,1,2,3,4], dtype='int64'))
         with localconverter(default_converter + rpyp.converter) as cv:
             rp_s = robjects.conversion.converter_ctx.get().py2rpy(s)
         # segfault before the fix
@@ -234,13 +250,13 @@ class TestPandasConversions(object):
         assert isinstance(rp_s, rinterface.StrSexpVector)
 
     def test_object2String(self):
-        series = pandas.Series(["a","b","c","a"], dtype="O")
+        series = pandas.Series(['a', 'b', 'c', 'a'], dtype='O')
         with localconverter(default_converter + rpyp.converter) as cv:
             rp_c = robjects.conversion.converter_ctx.get().py2rpy(series)
             assert isinstance(rp_c, rinterface.StrSexpVector)
 
     def test_object2String_with_None(self):
-        series = pandas.Series([None, "a","b","c","a"], dtype="O")
+        series = pandas.Series([None, 'a', 'b', 'c', 'a'], dtype='O')
         with localconverter(default_converter + rpyp.converter) as cv:
             rp_c = robjects.conversion.converter_ctx.get().py2rpy(series)
             assert isinstance(rp_c, rinterface.StrSexpVector)
@@ -392,27 +408,103 @@ class TestPandasConversions(object):
         s = s.split('\n')
         assert repr_str == s[2].strip()
 
-    def test_ri2pandas(self):
-        rdataf = robjects.r('data.frame(a=1:2, '
-                            '           b=I(c("a", "b")), '
-                            '           c=c("a", "b"))')
+    @pytest.mark.parametrize(
+        'colnames',
+        (('w', 'x', 'y', 'z'),
+         ('w', 'x', 'y', 'x'))
+    )
+    @pytest.mark.parametrize(
+        'r_rownames,py_rownames',
+        (('NULL', ('1', '2')),
+         ('c("a", "b")',('a', 'b')))
+    )
+    def test_ri2pandas(self, r_rownames, py_rownames, colnames):
+        rdataf = robjects.r(
+            f'data.frame({colnames[0]}=1:2, '
+            f'           {colnames[1]}=1:2, '
+            f'           {colnames[2]}=I(c("a", "b")), '
+            f'           {colnames[3]}=factor(c("a", "b")), '
+            f'           row.names={r_rownames}, '
+            '            check.names=FALSE)')
         with localconverter(default_converter + rpyp.converter) as cv:
             pandas_df = robjects.conversion.converter_ctx.get().rpy2py(rdataf)
 
         assert isinstance(pandas_df, pandas.DataFrame)
-        assert ('a', 'b', 'c') == tuple(pandas_df.keys())
-        assert pandas_df['a'].dtype in (numpy.dtype('int32'),
+        assert colnames == tuple(pandas_df.keys())
+        assert pandas_df['w'].dtype in (numpy.dtype('int32'),
                                         numpy.dtype('int64'))
-        assert pandas_df['b'].dtype == numpy.dtype('O')
-        assert isinstance(pandas_df['c'].dtype,
-                          pandas.api.types.CategoricalDtype)
+        assert pandas_df['y'].dtype == numpy.dtype('O')
+        if 'z' in colnames:
+            assert isinstance(pandas_df['z'].dtype,
+                              pandas.api.types.CategoricalDtype)
+        assert tuple(pandas_df.index) == py_rownames
 
-    def test_ri2pandas(self):
-        rdataf = robjects.r('data.frame(a=1:2, '
-                            '           row.names=c("a", "b"))')
+    @pytest.mark.parametrize(
+        'rcode,names,values',
+        (
+            ('list(list(x=1, y=3), list(x=2, y=4))',
+             ('x', 'y'), ((1, 2), (3, 4))),
+            ('list(list(x=1), list(x=2, y=4))',
+             ('x', 'y'), ((1, 2), (None, 4)))
+        )
+    )
+    def test__records_to_columns(self, rcode, names, values):
+        rlist = robjects.r(rcode)
+        columns = rpyp._records_to_columns(rlist)
+        assert tuple(columns.keys()) == names
+        for n, v in zip(names, values):
+            assert tuple(columns[n]) == v
+
+    @pytest.mark.parametrize(
+        'rcode,names,values',
+        (
+            ('data.frame('
+             '  a = c("a", "b"),'
+             '  b = c(1, 2)'
+             ')',
+             ('a', 'b'),
+             (["a", "b"], [1, 2])),
+            ('data.frame('
+             '  a = c("a", "b"),'
+             '  b = I(list(list(x=1, y=3), list(x=2, y=4)))'
+             ')',
+             ('a', ('b', 'x'), ('b', 'y')),
+             (["a", "b"], [1, 2], [3, 4]))
+        )
+    )
+    def test__flatten_dataframe(self, rcode, names, values):
+        rdataf = robjects.r(rcode)
+        colnames_lst = []
+        columns = tuple(rpyp._flatten_dataframe(rdataf, colnames_lst))
+        assert tuple(colnames_lst) == names
+
+    @pytest.mark.parametrize(
+        'rcode,names,values',
+        (
+            ('data.frame('
+             '  a = c("a", "b"),'
+             '  b = I(list(list(x=1, y=3), list(x=2, y=4)))'
+             ')',
+             ('a', ('b', 'x'), ('b', 'y')),
+             (["a", "b"], [1, 2], [3, 4])),
+            ('data.frame('
+             '  b = I(list(list(x=1, y=3), list(x=2, y=4)))'
+             ')',
+             (('b', 'x'), ('b', 'y')),
+             ([1, 2], [3, 4])),
+            ('aggregate(gear ~ am, mtcars, '
+             'function(x) c(mean = mean(x), sd = sd(x)), '
+             'simplify=FALSE)',
+             ('am', ('gear', 'mean'), ('gear', 'sd')),
+             ([0.0, 1.0], [3.210526, 4.384615],
+              [0.418854, 0.506370]))
+        )
+    )
+    def test_dataframe_with_list(self, rcode, names, values):
+        rdataf = robjects.r(rcode)
         with localconverter(default_converter + rpyp.converter) as cv:
-            pandas_df = cv.rpy2py(rdataf)
-        assert all(x == y for x, y in zip(rdataf.rownames, pandas_df.index))
+            pandas_df = robjects.conversion.converter_ctx.get().rpy2py(rdataf)
+        assert tuple(pandas_df.columns) == names
 
     def test_ri2pandas_issue207(self):
         d = robjects.DataFrame({'x': 1})
